/**
/*   Copyright (c) 2003by  Marco Welti
/*
/*   This document is  bound by the  QT Public License
/*   (http://www.trolltech.com/licenses/qpl.html).
/*   See License.txt for more information.
/*
/*
/*
/*   ALL RIGHTS RESERVED.  
/* 
/*   THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS
/*   OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
/*   WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
/*   ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY
/*   DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
/*   DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
/*   GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
/*   INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
/*   WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
/*   NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
/*   SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
/* 
/***********************************************************************/

#include "stdafx.h"

#include "CTagsHierarchyLoader.h"

#include <Model/LoaderBroker.h>
#include <Model/CTags/CTagsDatabase.h>
#include <Model/CTags/CTagsMetaObjectFactory.h>

#include <list>

using namespace std;

namespace CTags
{

struct MultipleDefinedSymbolException : public std::runtime_error
{
  MultipleDefinedSymbolException(const std::string &what) 
  : runtime_error(what) {}
}; 

struct UndefinedSymbolException : public std::runtime_error
{
  UndefinedSymbolException(const std::string &what) 
  : runtime_error(what) {}
};


Inheritable& HierarchyLoader::loadHierarchy(Inheritable &type, bool recursive)
{
  type.setExtendors(loadSubtypeHierarchy(type, recursive));
  type.setParents(loadSupertypeHierarchy(type, recursive));
  return type;
}

std::list<smart_ptr<Inheritable> > HierarchyLoader::loadSubtypeHierarchy(const Inheritable &type, bool recursive)
{
  return getSubHierarchy(type.getFQN(), recursive);
}

std::list<smart_ptr<Inheritable> > HierarchyLoader::loadSupertypeHierarchy(const Inheritable &type, bool recursive)
{
  return getSuperHierarchy(type.getFQN(), recursive);
}

std::list<smart_ptr<Inheritable> > HierarchyLoader::getSuperHierarchy(const std::string &fqn, bool recursive)
{
  Tag tag;
  try {
    tag = getTag(fqn);
    
  } catch(UndefinedSymbolException &) { }

  std::list<smart_ptr<Inheritable> > parents;
  for(list<string>::const_iterator parent = tag.getInheritance().begin(); parent != tag.getInheritance().end(); ++parent) {
    smart_ptr<Inheritable> super;
    try {
      super = makeRootNode(getTag(*parent));
      if(recursive) {
        super->setParents(getSuperHierarchy(*parent, recursive));
      }
      
    
    } catch(UndefinedSymbolException &) {
      super = makeUnknown(*parent);
    }
    parents.push_back(super);
  }

  return parents;
}

struct InheritsIs : public std::binary_function<Tag, string, bool>
{
  result_type operator()(const first_argument_type &tag, const second_argument_type &name) const
  {
    return find(tag.getInheritance().begin(), tag.getInheritance().end(), name) != tag.getInheritance().end();
  }
};

std::list<smart_ptr<Inheritable> > HierarchyLoader::getSubHierarchy(const std::string &fqn, bool recursive)
{

  list<Tag> children = getDatabase().selectTagsWhere(make_query(InheritsIs(), fqn));
  std::list<smart_ptr<Inheritable> > extendors;
  
  for(list<Tag>::iterator child = children.begin(); child != children.end(); ++child) {
    smart_ptr<Inheritable> extendor = makeRootNode(*child);
    if(recursive) {
      extendor->setExtendors(getSubHierarchy(child->getFQN(), recursive));
    }
    extendors.push_back(extendor);
  }
  
  return extendors;
}

Tag HierarchyLoader::getTag(const std::string &fqn)
{
  list<Tag> ctags = getDatabase().selectTagsWhere(make_query(FullQualifiedNameIs(), fqn));
 
  if(ctags.empty()) {
    throw UndefinedSymbolException(fqn);

  } else if (ctags.size() > 1) {
    stringstream ss;
    ss << "tag-count " << ctags.size() << "!=1 for tag " << fqn << endl << endl << "found declarations at" << endl;
    for(list<Tag>::iterator i = ctags.begin(); i != ctags.end(); i++) {
      ss << i->getFQN() << " " << i->getFile() << ":"<< i->getLine() << endl;
    }
    throw MultipleDefinedSymbolException(ss.str());
  }
  
  return ctags.front();
}

smart_ptr<Inheritable> HierarchyLoader::makeRootNode(const Tag &tag)
{
  smart_ptr<CppType> inheritable = TypeFactory().createTag(tag);
  inheritable->setBroker(LoaderBroker::instance(mWorkspace));
  return dynamic__cast<Inheritable>(inheritable);
}

smart_ptr<Inheritable> HierarchyLoader::makeUnknown(const std::string &fqn)
{
  smart_ptr<Inheritable> unknown = make_smart_ptr(new Class(fqn, ePublic, TagLocation()));
  unknown->setBroker(LoaderBroker::instance(mWorkspace));
  return unknown;
}

Database& HierarchyLoader::getDatabase()
{
  return *DatabaseManager::instance().getDatabase(mWorkspace);
}

}
